package geniusweb.protocol.session.mopac2;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.junit.Before;
import org.junit.Test;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import geniusweb.actions.Action;
import geniusweb.actions.PartyId;
import geniusweb.progress.Progress;
import geniusweb.progress.ProgressTime;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.session.TeamInfo;
import geniusweb.protocol.session.mopac2.phase.OfferPhase;
import geniusweb.protocol.session.mopac2.phase.Phase;
import geniusweb.references.Parameters;
import geniusweb.references.PartyWithParameters;
import geniusweb.references.PartyWithProfile;
import geniusweb.references.ProfileRef;
import geniusweb.voting.evaluatorwithvalue.LargestAgreementWithValue;
import tudelft.utilities.junit.GeneralTests;

public class MOPAC2StateTest extends GeneralTests<MOPAC2State> {
	private static final Date NOW = new Date(1001);

	private final ObjectMapper jackson = new ObjectMapper();

	private final MOPAC2Settings settings = mock(MOPAC2Settings.class);

	private final MOPAC2State state = new MOPAC2State(settings);

	private PartyId PARTY1 = new PartyId("party1");
	private PartyId PARTY2 = new PartyId("party2");
	private PartyId PARTY3 = new PartyId("party3");

	private PartyWithParameters partyparams1 = mock(PartyWithParameters.class);
	private Parameters params1 = new Parameters();

	private PartyWithProfile profile1 = new PartyWithProfile(partyparams1,
			mock(ProfileRef.class));

	// settings is copy from the MOPAC2Settings string. Shortcut to creating
	// this all manually.
	private final String mopacsettingstr = "{\"MOPAC2Settings\":"
			+ "{\"participants\":["
			+ "{\"TeamInfo\":{\"parties\":[ {\"party\":{\"partyref\":\"http://party1\",\"parameters\":{}},\"profile\":\"http://profile1\"} ]}},"
			+ "{\"TeamInfo\":{\"parties\":[ {\"party\":{\"partyref\":\"http://party2\",\"parameters\":{}},\"profile\":\"http://profile2\"} ]}}  ],"
			+ "\"deadline\":{\"DeadlineTime\":{\"durationms\":100}},\"votingevaluator\":{\"LargestAgreementWithValue\":{}}}}";;
	// create real objects for the GeneralTests and for serialization tests
	private PartyStates partyStates = new PartyStates(Collections.emptyMap());
	private Phase phase1 = new OfferPhase(partyStates, 100l,
			new LargestAgreementWithValue());
	private Phase phase2 = new OfferPhase(partyStates, 200l,
			new LargestAgreementWithValue());
	private Progress progress1 = new ProgressTime(100l, NOW);
	private Progress progress2 = new ProgressTime(200l, NOW);

	private MOPAC2State state1, state1a, state2, state3, state4, state5, state6;

	private String serialized = "{\"MOPAC2State\":{\"phase\":{\"OfferPhase\":{\"partyStates\":{\"notYetActed\":[],\"actions\":[],\"agreements\":{},\"walkedAway\":[],\"exceptions\":{},\"powers\":{}},\"evaluator\":{\"LargestAgreementWithValue\":{}},\"deadline\":100}},\"actions\":[],\"progress\":{\"ProgressTime\":{\"duration\":100,\"start\":1001}},\"settings\":{\"MOPAC2Settings\":{\"participants\":[{\"TeamInfo\":{\"parties\":[{\"party\":{\"partyref\":\"http://party1\",\"parameters\":{}},\"profile\":\"http://profile1\"}]}},{\"TeamInfo\":{\"parties\":[{\"party\":{\"partyref\":\"http://party2\",\"parameters\":{}},\"profile\":\"http://profile2\"}]}}],\"deadline\":{\"DeadlineTime\":{\"durationms\":100}},\"votingevaluator\":{\"LargestAgreementWithValue\":{}}}},\"partyprofiles\":{}}}";

	@Before
	public void before()
			throws JsonParseException, JsonMappingException, IOException {
		params1 = params1.with("power", 2);
		when(partyparams1.getParameters()).thenReturn(params1);

		MOPAC2Settings realsettings = jackson.readValue(mopacsettingstr,
				MOPAC2Settings.class);
		TeamInfo party = mock(TeamInfo.class);
		PartyWithParameters pwithp = mock(PartyWithParameters.class);
		when(pwithp.getParameters()).thenReturn(new Parameters());
		when(party.getParties()).thenReturn(Arrays
				.asList(new PartyWithProfile(pwithp, mock(ProfileRef.class))));
		when(party.getSize()).thenReturn(1);
		MOPAC2Settings realsettings2 = realsettings.with(party);

		state1 = new MOPAC2State(phase1, Collections.emptyList(), progress1,
				realsettings, Collections.emptyMap());
		state1a = new MOPAC2State(phase1, Collections.emptyList(), progress1,
				realsettings, Collections.emptyMap());
		state2 = new MOPAC2State(phase2, Collections.emptyList(), progress1,
				realsettings, Collections.emptyMap());
		state3 = new MOPAC2State(phase1, Arrays.asList(mock(Action.class)),
				progress1, realsettings, Collections.emptyMap());
		state4 = new MOPAC2State(phase1, Collections.emptyList(), progress2,
				realsettings, Collections.emptyMap());
		state5 = new MOPAC2State(phase1, Collections.emptyList(), progress1,
				realsettings2, Collections.emptyMap());
		state6 = new MOPAC2State(phase1, Collections.emptyList(), progress1,
				realsettings,
				Collections.singletonMap(PARTY1, mock(PartyWithProfile.class)));

	}

	@Override
	public List<List<MOPAC2State>> getGeneralTestData() {
		return Arrays.asList(Arrays.asList(state1, state1a),
				Arrays.asList(state2), Arrays.asList(state3),
				Arrays.asList(state4), Arrays.asList(state5),
				Arrays.asList(state6));
	}

	@Override
	public List<String> getGeneralTestStrings() {
		return Arrays.asList(
				"MOPAC2State.*OfferPhase.*PartyStates.*,MOPAC2Settings.*PartyWithProfile.*ProgressTime.*",
				"MOPAC2State.*OfferPhase.*PartyStates.*,MOPAC2Settings.*PartyWithProfile.*ProgressTime.*",
				"MOPAC2State.*OfferPhase.*PartyStates.*,MOPAC2Settings.*PartyWithProfile.*ProgressTime.*",
				"MOPAC2State.*OfferPhase.*PartyStates.*,MOPAC2Settings.*PartyWithProfile.*ProgressTime.*",
				"MOPAC2State.*OfferPhase.*PartyStates.*,MOPAC2Settings.*PartyWithProfile.*ProgressTime.*",
				"MOPAC2State.*OfferPhase.*PartyStates.*,MOPAC2Settings.*PartyWithProfile.*ProgressTime.*");
	}

	@Test
	public void smokeTest() {
	}

	@Test
	public void addProfilesBeforeInit() {
		MOPAC2State st = state.with(PARTY1, profile1);
		assertTrue(st.getPartyProfiles().containsKey(PARTY1));
	}

	@Test
	public void testInit() {
		// no parties were added, should throw? Go to final mode?
		// Notice that it's possible to create settings with 0 participants,
		// because this is needed for Tournament mechanism.
		MOPAC2State state1 = state
				.initPhase(new ProgressTime(100l, new Date(0)), 1);
		assertTrue(state1.getActions().isEmpty());
		assertTrue(state1.getAgreements().getMap().isEmpty());
		assertTrue(state1.getPartyProfiles().isEmpty());
		Phase phase1 = state1.getPhase();
		assertTrue(phase1 instanceof OfferPhase);
		OfferPhase offer1 = (OfferPhase) phase1;
		assertEquals(new PartyStates(Collections.emptyMap()),
				offer1.getPartyStates());
	}

	@Test
	public void getAgreementsTest() {
		MOPAC2State st = state.initPhase(new ProgressTime(100l, new Date(0)),
				1);
		assertTrue(st.getAgreements().getMap().isEmpty());
	}

	@Test
	public void testDeserialize() throws IOException {
		MOPAC2State obj = jackson.readValue(serialized, MOPAC2State.class);
		System.out.println(obj);
		assertEquals(state1, obj);
	}

	@Test
	public void testSerialize()
			throws JsonProcessingException, URISyntaxException {

		String string = jackson.writeValueAsString(state1);
		System.out.println(string);
		assertEquals(serialized, string);
	}

	@Test
	public void testNewPhasePossible() {

		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		when(progress.getTerminationTime())
				.thenReturn(new Date(Phase.PHASE_MAXTIME)); // plenty for new
															// phase
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPAC2State s1 = new MOPAC2State(phase, actions, progress, settings,
				partyprofiles);

		PartyStates states = mock(PartyStates.class);
		when(phase.getPartyStates()).thenReturn(states);
		when(states.getNegotiatingParties())
				.thenReturn(new HashSet<>(Arrays.asList(PARTY1, PARTY2)));
		assertTrue(s1.isNewPhasePossible(10));
	}

	@Test
	public void testNewPhasePossibleOnly1Party() {

		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		when(progress.getTerminationTime())
				.thenReturn(new Date(Phase.PHASE_MAXTIME)); // plenty for new
															// phase
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPAC2State s1 = new MOPAC2State(phase, actions, progress, settings,
				partyprofiles);

		PartyStates states = mock(PartyStates.class);
		when(phase.getPartyStates()).thenReturn(states);
		when(states.getNegotiatingParties())
				.thenReturn(new HashSet<>(Arrays.asList(PARTY1)));
		assertFalse(s1.isNewPhasePossible(10));
	}

	@Test
	public void testNewPhasePossibleNotEnoughTime() {

		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		when(progress.getTerminationTime())
				.thenReturn(new Date(Phase.PHASE_MINTIME - 10));
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPAC2State s1 = new MOPAC2State(phase, actions, progress, settings,
				partyprofiles);

		PartyStates states = mock(PartyStates.class);
		when(phase.getPartyStates()).thenReturn(states);
		when(states.getNegotiatingParties())
				.thenReturn(new HashSet<>(Arrays.asList(PARTY1)));
		assertFalse(s1.isNewPhasePossible(10));
	}

	@Test
	public void nextPhaseTest() {
		long NOW = 102;
		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		when(progress.getTerminationTime())
				.thenReturn(new Date(Phase.PHASE_MAXTIME));
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPAC2State s1 = new MOPAC2State(phase, actions, progress, settings,
				partyprofiles);

		// PhaseB is the next phase, with remaining time
		Phase phaseB = mock(Phase.class);
		when(phase.next(eq(NOW), eq(Phase.PHASE_MAXTIME - NOW)))
				.thenReturn(phaseB);

		MOPAC2State s2 = s1.nextPhase(NOW);
		assertEquals(phaseB, s2.getPhase());
	}

	@Test
	public void exceptionTest() {
		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPAC2State s1 = new MOPAC2State(phase, actions, progress, settings,
				partyprofiles);

		Phase exphase = mock(Phase.class); // supposed phase that contains the
											// exception
		when(phase.with(any(ProtocolException.class))).thenReturn(exphase);

		MOPAC2State errstate = s1.with(new ProtocolException("test", PARTY1));
		assertEquals(exphase, errstate.getPhase());
		assertEquals(progress, errstate.getProgress());

	}

}
